type proc_viewer = {
    frame	: Widget.frame Widget.widget;
    text  	: Widget.text Widget.widget;
  }

type t = proc_viewer


let create  frame =
  let f = Frame.create frame ~borderwidth:10 in
  let s = "PROCESS LIST\n" ^ (Printf.sprintf "%20s | %8s | %5s | %s" "command" "user" "CPU" "hosts") in
  let l = Label.create f ~text:s and 
      t = Text.create f ~height:60 ~width:70 ~state:`Disabled in
  begin
    Tk.pack ~side:`Bottom [t];
    Tk.pack ~side:`Top [l];
     { frame = f;
       text  = t; }
  end;;


let delete_all_text  t =
   let start = (`Linechar(1,0), []) and
       stop = (`End, []) in
   Text.delete t ~start:start ~stop:stop;;


let create_tag  t line_no tag_name =
    let tag  = tag_name and
	strt = `Linechar (line_no, 0), [] and
	stp  = `Linechar (line_no+1, 0), [] in 
  let _ = Text.tag_add t ~tag:tag_name ~start:strt ~stop:stp in
  tag;;





type proc_stat = 
    { comm 	: string;
      user	: string;
      uptime	: float * float; (* user and sys *)
      hosts	: string list
   }
      

let calc_uptime_sub  old_ptable hostname interval x =
    let uptime = 
	try 
	    let y = Ptable.find x.Proc.pid old_ptable in
	    let f x1 x2 =
		let b = (float_of_int (x1 - x2)) /. interval in
		if b > 100.0 then 100.0 else b
	    in
	    (f x.Proc.utime y.Proc.utime, f x.Proc.stime y.Proc.stime)
	with _ -> (0.0, 0.0)
    in
    { comm   = x.Proc.comm;
      user   = x.Proc.user;
      uptime = uptime;
      hosts  = [hostname] };;
    

let calc_uptime  old_x new_x =
    let (_, (old_stat, old_time, _))        = old_x and
	(hostname, (new_stat, new_time, _)) = new_x in
    let old_ptable = old_stat.Stat.ptable and
	new_ptable = new_stat.Stat.ptable in
    let l = Ptable.to_vals new_ptable in
    let interval = new_time -. old_time in 
    List.map (calc_uptime_sub old_ptable hostname interval) l;;
    

(*********************************************************************************)
  

let merge_hostnames l =
    let l' = List.map (function x -> x.hosts) l in
    let l'' = List.flatten l' in
    let l''' = MyList.filter_dup String.compare l'' in
    List.fast_sort String.compare l''';;


let rec merge_sub  retval = function
    [] -> retval
  | proc :: _ as l ->
      let (l', rest) = List.partition (function x -> (x.comm = proc.comm) && (x.user = proc.user)) l in (* extract the same cmdline *)
      let uptime =
 	  let (u, s) = List.fold_left (fun (u, s) x -> (u +. (fst x.uptime), s +. (snd x.uptime))) (0.0, 0.0) l' in
	  let f x = 
	      x /. (float_of_int (List.length l')) in
   (*  (f u, f s)  *)
	  let x  = List.hd l' in
	  (fst x.uptime, snd x.uptime)
      in
      let hosts = merge_hostnames l' in
      let proc' = { comm = proc.comm;
		    user  = proc.user;
		    uptime = uptime;
		    hosts  = hosts }
      in
      let retval' = proc' :: retval in
      merge_sub  retval' rest;;


let merge  l =
    let l' = List.flatten l in
    merge_sub [] l';;


(*********************************************************************************)


let filter = 
    let f x = 
	let u = x.uptime in 
	let u' = (fst u) +. (snd u) in
	(u' > 0.0) || (List.mem x.user Vconfig.users)
    in
    List.filter f;;


(*********************************************************************************)


let sort = 
    let f x y = 
	let g x = let u = x.uptime in (fst u) +. (snd u) in
	int_of_float ((g y) -. (g x))
    in
    List.fast_sort f;;


(*********************************************************************************)


let extract_prefix  s = 
    let regexp = Str.regexp "[^0-9]*" in
    let _ = Str.search_forward regexp s 0 in
    Str.matched_string s;;


let prefix_is_eq  prefix s = try
    let n = String.length prefix in
    let s' = String.sub s 0 n in
    s' = prefix
with _ -> false;;


let extract_suffix  prefix s = 
    let n = String.length prefix and
	m = String.length s in
    String.sub s n (m - n);;


let suffix_is_succ  s1 s2 =
    let n1 = int_of_string s1 and
	n2 = int_of_string s2 in
    n2 = n1 + 1;;


let rec merge_suffixes_sub  range = function
    [] -> (range, [])
  | h :: t as l ->
      let (lower, upper) = range in
      if suffix_is_succ upper h
      then merge_suffixes_sub (lower, h) t
      else (range, l);;


let merge_suffixes  retval (h, t) =
    let (s, rest) = merge_suffixes_sub (h, h) t in
    (s :: retval, rest);;


let suffix_range_to_string  (lower, upper) =
    if lower = upper 
    then lower 
    else "[" ^ lower ^ "-" ^ upper ^ "]";;


let extract_suffixes  prefix l =
    let l' = List.map (extract_suffix prefix) l in
    let l'' = MyList.fold_left merge_suffixes [] l' in
    let l''' = List.rev l'' in
    MyList.to_string suffix_range_to_string "," l''';;


let hosts_to_string_sub  retval (h, t) =
    let prefix = extract_prefix h in
    let (l', rest) = List.partition (prefix_is_eq prefix) (h :: t) in
    let suffixes = extract_suffixes prefix l' in
    let retval' = (prefix, suffixes) :: retval in
    (retval', rest);;


let hosts_to_string  l =
    let l' = MyList.fold_left hosts_to_string_sub [] l in
    let l'' = List.rev l' in
    let f (prefix, suffixes)  = prefix ^ suffixes in
    MyList.to_string f ", " l'';;


let to_string  x =
    let u = let u = x.uptime in (fst u) +. (snd u) in
    let s =
	let comm = x.comm and
	    user = x.user and
	    uptime = u and
	    hosts = hosts_to_string x.hosts in
	Printf.sprintf "%20s | %8s | %3.1f%% | %s \n" comm user uptime hosts 
    in
    let color = if u > 10.0 then `Red else `Black in
    (s, color);;


(*********************************************************************************)


let rec draw_sub  text line_no = function
    [] -> ()
  | h :: t ->
      let (s, color) = h in
      let () = Text.insert  ~index:(`Linechar(line_no,0), []) ~text:s text in
      let () = 
	  let tag = create_tag  text line_no s in
	  Text.tag_configure  text ~tag:tag ~foreground:color 
      in
      draw_sub  text (line_no + 1) t;;


let draw  x old_stable stable = 
    let l = 
	let rec g result old_l new_l =
	    match old_l with
	      [] -> result
	    | old_h :: old_t ->
		match new_l with
		  [] -> result
		| new_h :: new_t ->
		    let n = String.compare (fst old_h) (fst new_h) in
		    if n = 0
		    then
			let result' = (calc_uptime old_h new_h) :: result in
			g result' old_t new_t
		    else if n > 0
		    then g result old_l new_t
		    else g result old_t new_l
	in
	g [] (Stable.to_list old_stable) (Stable.to_list stable)
    in
    let l' = 
	let l'   = merge l in
	let l''  = filter l' in
	let l''' = sort l'' in
	List.map to_string l'''
    in
    let t = x.text in
    begin
	Text.configure t ~state:`Normal;
	delete_all_text t;
	draw_sub t 1 l';
	Text.configure t ~state:`Disabled
    end;;
